import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;

public class Application {
	
	static int PROBES_BOUND = -1;   // upper bound on #probes expected in data 
	static int PROBE_SIZE = -1;	 	// size of probes in data 
	static int MER_SIZE = -1;		// size of k-mers to work with 
	static int MIN_PROBES = -1;	 	// minimum number of probes where k-mer must appear 	
	static int MAX_KMERS = -1;		// size of "thickest" layer of top k-mers to be considered when computing the motif  		 
	static int MID_KMERS = -1;		// size of "middle" layer of top k-mers to be considered when computing the motif 
	static int MIN_KMERS = -1;		// size of "thinnest" layer of top k-mers to be considered when computing the motif
	static int MAX_MOTIF = -1;		// upper bound on motif size
	static int PROBE_CUTOFF = -1; 	// #probes to consider per k-mer
	static float CORE_THRESHOLD = -1; 	 // degree of preservation considered core
	static float ENTROPY_THRESHOLD = -1; // max entropy for significant position
	static int MAX_OFFSET = -1;			 // max allowable offset of a k-mer aligned to the center string
	static float MIN_MATCHES = -1; 		 // min allowable #matches for valid alignment (after decimal point indicates #consecutive matches)
	
	/* Weighting scheme for mixing max, mid and min PWMs in core (first) positions, next
	   to core (second) positions, and positions further away (third) 
	*/ 
	static float MAX_FIRST_FRAC = -1; static float MAX_SECOND_FRAC = -1; static float MAX_THIRD_FRAC = -1;
	static float MID_FIRST_FRAC = -1; static float MID_SECOND_FRAC = -1; static float MID_THIRD_FRAC = -1;
	static float MIN_FIRST_FRAC = -1; static float MIN_SECOND_FRAC = -1; static float MIN_THIRD_FRAC = -1;

	/**
	 * @param  Program runs with 4 arguments: parameters filename, input probes-and-scores filename, output PWM filename and logo flag (-d for display; -n for no display)
	 * @throws IOException in case of file read/write failures, and terminates responsibly. 
	 */
	public static void main(String[] args) throws IOException {
			
		if(args.length != 3) {
			System.out.println("Usage: <parameters filename> <probes and scores filename> <output PWM filename>");
			return;
		}
		
		MAX_FIRST_FRAC = MAX_SECOND_FRAC = MAX_THIRD_FRAC = (float)0.34;
		MID_FIRST_FRAC = MID_SECOND_FRAC = MID_THIRD_FRAC = (float)0.33;
		MIN_FIRST_FRAC = MIN_SECOND_FRAC = MIN_THIRD_FRAC = (float)0.33;
		MID_KMERS = 500;
		MIN_KMERS = 500;
		CORE_THRESHOLD = 1;
		
		/* Read parameters file */
		BufferedReader in1 = null;
		try {
			in1 = new BufferedReader(new FileReader(args[0]));
			while(in1.ready()) {
				String line = in1.readLine();
				String[] arr = line.split("\t");
				if(arr[0].equals("PROBES_BOUND")) { PROBES_BOUND = Integer.parseInt(arr[1]); continue;}
				if(arr[0].equals("PROBE_SIZE")) { PROBE_SIZE = Integer.parseInt(arr[1]); continue;}
				if(arr[0].equals("MER_SIZE")) { MER_SIZE = Integer.parseInt(arr[1]); continue;}
				if(arr[0].equals("MIN_PROBES")) { MIN_PROBES = Integer.parseInt(arr[1]); continue;}
				if(arr[0].equals("MAX_KMERS")) { MAX_KMERS = Integer.parseInt(arr[1]); continue;}
				if(arr[0].equals("MAX_MOTIF")) { MAX_MOTIF = Integer.parseInt(arr[1]); continue;}
				if(arr[0].equals("PROBE_CUTOFF")) { PROBE_CUTOFF = Integer.parseInt(arr[1]); continue;}
				if(arr[0].equals("MAX_OFFSET")) { MAX_OFFSET = Integer.parseInt(arr[1]); continue;}
				if(arr[0].equals("ENTROPY_THRESHOLD")) { ENTROPY_THRESHOLD = Float.parseFloat(arr[1]); continue;}
				if(arr[0].equals("MIN_MATCHES")) { MIN_MATCHES = Float.parseFloat(arr[1]); continue;}
			}
			in1.close(); // release resource 
		} catch (IOException e) {
			e.printStackTrace();
			return;
		}
		if(!validateParameters()) { return; } // elementary validation of parameters
		
		/* Initialize k-mer objects */
		int numKMers = (int)Math.pow(4, MER_SIZE);
		MerObj[] kMers = new MerObj[numKMers]; 		 // will hold k-mer objects 
		Integer[] kMersSort = new Integer[numKMers]; // will hold Integer objects sorted by the score of the corresponding k-mer 
		enumerateMers(kMers);	
		
		/* Read data-set from file, breaking probes into k-mers as they are read (to save time) */
		char[][] probes = new char[PROBES_BOUND][PROBE_SIZE]; 
		float[] scores = new float[PROBES_BOUND];
		int probeIndex = 0;
		
		int mer = 0;    // will hold the integer representation of each formed k-mer
		int picker = 0; // will denote the rightmost (MER_SIZE*2) bits of a k-mer integer representation
		for (int i = 0 ; i < MER_SIZE ; i++) {
			picker <<= 2;
			picker |= 0x3;
		}
		
		BufferedReader in2 = null;
		try {
			in2 = new BufferedReader(new FileReader(args[1]));
			while(in2.ready()) {
				String line = in2.readLine();
				String[] arr = line.split("\t");
				probes[probeIndex] = (arr[1].substring(0, 36)).toCharArray(); // arr[0].toCharArray();
				scores[probeIndex] = Float.valueOf(arr[0]).floatValue(); // Float.valueOf(arr[1]).floatValue();
				for (int i = 0 ; i < PROBE_SIZE ; i++) {
					switch (probes[probeIndex][i]) {
						// case 'A' = 00 -> therefore not needed
						case 'C': { mer |= 0x1 ; break; } // 'C' = 01
						case 'G': { mer |= 0x2 ; break; } // 'G' = 10
						case 'T': { mer |= 0x3 ; break; } // 'T' = 11
					}
					mer &= picker; // take only rightmost (MER_SIZE*2) bits
					if (i >= MER_SIZE-1) {
						kMers[mer].addInstance(probeIndex+1, i-(MER_SIZE-1), scores[probeIndex]); // remember the rank and offset in probe inside the k-mer object
					}
					mer <<= 2;
				}
				probeIndex++;
			}
			in2.close(); // release resource 
		} catch (FileNotFoundException e) {
			e.printStackTrace();
			return;
		} 
						
		/* Compute WMW and average binding scores */
		for (int i = 0; i < kMers.length ; i++) {
			kMersSort[i] = (Integer)i;
			if (kMers[i].getForegroundSize() < MIN_PROBES) { // if k-mer occurrence is abnormally low, this is a "bug" in the experimental data and we weed it out	  
				kMers[i] = null;
			}
			else { kMers[i].computeWMWAndAvg(probeIndex); }
		}
		
		/* Sort kMersSort array using an instance of the RankCompare comparator, with flag 0 for WMW and 1 for avg. binding score */ 
		java.util.Arrays.sort(kMersSort, new RankCompare(kMers, 1));
		
		/* Align MAX_KMERS top ranking k-mers, instantiating an MSAMotif object */
		MSAMotif topKMers = new MSAMotif(kMers, kMersSort, MAX_OFFSET, MIN_MATCHES);

		/* Request the motif PWM from the MSAMotif object */
		double[][] PWM = topKMers.getMotifPWM(probes, scores);
		
		/* Print PWM to file */  
		BufferedWriter out = null;
		try {
			out = new BufferedWriter(new FileWriter(new File(args[2])));
			out.write("A:\t");
			for (int i = 0; i < PWM.length; i++) {
				out.write(PWM[i][0] + "\t");
			}
			out.write("\nC:\t");
			for (int i = 0; i < PWM.length; i++) {
				out.write(PWM[i][1] + "\t");
			}
			out.write("\nG:\t");
			for (int i = 0; i < PWM.length; i++) {
				out.write(PWM[i][2] + "\t");
			}
			out.write("\nT:\t");
			for (int i = 0; i < PWM.length; i++) {
				out.write(PWM[i][3] + "\t");
			}
			out.close(); // release resource	
		} catch (IOException e) {
			e.printStackTrace();
			return;
		} 
		
		return;
		
	}
	
/* ======================================UTILITY FUNCTIONS===================================== */
	
	/* Some basic sanity checks for input parameters */
	private static boolean validateParameters() {
		boolean flag = true;
		if(PROBES_BOUND < 1) { flag = false; }
		if(PROBE_SIZE < 1) { flag = false; }
		if(MER_SIZE < 1) { flag = false; }
		if(MIN_PROBES < 0) { flag = false; }
		if(MAX_KMERS < 1) { flag = false; }
		if(MAX_MOTIF < 1) { flag = false; }
		if(PROBE_CUTOFF < 1) { flag = false; }
		if(MAX_OFFSET < 0) { flag = false; }
		if(CORE_THRESHOLD < 0 || CORE_THRESHOLD > 1) { flag = false; }
		if(ENTROPY_THRESHOLD < 0 || ENTROPY_THRESHOLD > 2) { flag = false; }
		if(MIN_MATCHES < 1) { flag = false; }
		if(MAX_FIRST_FRAC < 0 || MAX_FIRST_FRAC > 1) { flag = false; }
		if(MID_FIRST_FRAC < 0 || MID_FIRST_FRAC > 1) { flag = false; }
		if(MIN_FIRST_FRAC < 0 || MIN_FIRST_FRAC > 1) { flag = false; }
		if(MAX_FIRST_FRAC + MID_FIRST_FRAC + MIN_FIRST_FRAC != (float)1) { flag = false; } 
		if(MAX_SECOND_FRAC < 0 || MAX_SECOND_FRAC > 1) { flag = false; }
		if(MID_SECOND_FRAC < 0 || MID_SECOND_FRAC > 1) { flag = false; }
		if(MIN_SECOND_FRAC < 0 || MIN_SECOND_FRAC > 1) { flag = false; }
		if(MAX_SECOND_FRAC + MID_SECOND_FRAC + MIN_SECOND_FRAC != (float)1) { flag = false; } 
		if(MAX_THIRD_FRAC < 0 || MAX_THIRD_FRAC > 1) { flag = false; }
		if(MID_THIRD_FRAC < 0 || MID_THIRD_FRAC > 1) { flag = false; }
		if(MIN_THIRD_FRAC < 0 || MIN_THIRD_FRAC > 1) { flag = false; }
		if(MAX_THIRD_FRAC + MID_THIRD_FRAC + MIN_THIRD_FRAC != (float)1) { flag = false; } 
		
		if(!flag) { System.out.println("Not all parameters were input correctly. Pls try again."); } 
		return flag;
	}

	/* Populate kMers array, generating each k-mer string in turn as in a "quaternary" counter */
	private static void enumerateMers(MerObj[] kMers) {
		char s[] = new char[MER_SIZE];
		for (int i = 0 ; i < s.length ; i++) { s[i] = 'A'; }
		kMers[0] = new MerObj(new String(s));
		for (int i = 1 ; i < kMers.length ; i++) {
			s[MER_SIZE-1] = nextBase(s[MER_SIZE-1]);
			for (int j = 1 ; j < MER_SIZE ; j++) {
				if (i % ((int)Math.pow(4, j)) == 0) { 
					s[MER_SIZE-1-j] = nextBase(s[MER_SIZE-1-j]);
				}
			}
			kMers[i] = new MerObj(new String(s));
		}
	}
	
	/* Retrieve the next nucleotide while enumerating k-mers */
	private static char nextBase(char base) {
		switch (base) {
			case ('A'): { return 'C'; }
			case ('C'): { return 'G'; }
			case ('G'): { return 'T'; }
			case ('T'): { return 'A'; }
		}
		return ('X');
	}
	
	/* Present motif logo */
} 
